(function() {
  var ARRAY, BUFFER, BytewiseCodec, Codec, DATE, DATE_POST_EPOCH, DATE_PRE_EPOCH, Errors, FALSE, FLOAT, FUNCTION, INTEGER, InvalidArgumentError, InvalidFormatError, MaxIntSize, MaxIntValue, MaxUInt32, MaxUInt32Count, MaxUInt32Size, NEGATIVE, NEGATIVE_FLOAT, NEGATIVE_INFINITY, NEGATIVE_INT, NULL, NUMBER, NUMBER_FLOAT, NUMBER_INT, NUMBER_NEGATIVE_INFINITY, NUMBER_POSITIVE_INFINITY, NotSupportedError, OBJECT, POSITIVE, POSITIVE_FLOAT, POSITIVE_INFINITY, POSITIVE_INT, REGEXP, STRING, TRUE, UNDEFINED, aliases, escapeString, intToHex, isArray, isBuffer, isDate, isFunction, isNumber, isObject, isRegExp, isString, padChar, register, toFixedInt, unescapeString;

  Errors = require('abstract-object/Error');

  Codec = module.exports = require('buffer-codec');

  isString = require("abstract-object/lib/util/isString");

  isNumber = require("abstract-object/lib/util/isNumber");

  isDate = require("abstract-object/lib/util/isDate");

  isRegExp = require("abstract-object/lib/util/isRegExp");

  isArray = require("abstract-object/lib/util/isArray");

  isFunction = require("abstract-object/lib/util/isFunction");

  isObject = require("abstract-object/lib/util/isObject");

  register = Codec.register;

  aliases = Codec.aliases;

  isBuffer = Buffer.isBuffer;

  escapeString = Codec.escapeString;

  unescapeString = Codec.unescapeString;

  InvalidFormatError = Errors.InvalidFormatError;

  InvalidArgumentError = Errors.InvalidArgumentError;

  NotSupportedError = Errors.NotSupportedError;

  MaxIntValue = 0x20000000000000;

  MaxIntSize = 14;

  MaxUInt32 = 0xffffffff;

  MaxUInt32Count = 0x100000000;

  MaxUInt32Size = 8;

  NULL = ' ';

  NUMBER = 'N';

  NEGATIVE_INFINITY = '0';

  NUMBER_NEGATIVE_INFINITY = NUMBER + NEGATIVE_INFINITY;

  NEGATIVE = '-';

  POSITIVE = '0';

  INTEGER = 'i';

  NUMBER_INT = NUMBER + INTEGER;

  NEGATIVE_INT = NUMBER_INT + NEGATIVE;

  POSITIVE_INT = NUMBER_INT + POSITIVE;

  DATE = 'D';

  DATE_PRE_EPOCH = DATE + NEGATIVE;

  DATE_POST_EPOCH = DATE + POSITIVE;

  FLOAT = 'f';

  NUMBER_FLOAT = NUMBER + FLOAT;

  NEGATIVE_FLOAT = NUMBER_FLOAT + NEGATIVE;

  POSITIVE_FLOAT = NUMBER_FLOAT + POSITIVE;

  POSITIVE_INFINITY = '9';

  NUMBER_POSITIVE_INFINITY = NUMBER + POSITIVE_INFINITY;

  BUFFER = 'B';

  STRING = '"';

  ARRAY = '[';

  OBJECT = '{';

  REGEXP = 'R';

  FUNCTION = 'f';

  FALSE = 'F';

  TRUE = 'T';

  UNDEFINED = '~';

  padChar = function(c, len) {
    var result;
    if (len == null) {
      len = 2;
    }
    result = '';
    len++;
    while (len -= 1) {
      result += c;
    }
    return result;
  };

  toFixedInt = function(value, digits, radix) {
    var result;
    if (digits == null) {
      digits = MaxUInt32Size;
    }
    if (radix == null) {
      radix = 16;
    }
    result = padChar(0, digits);
    if (value < 0) {
      value += MaxUInt32Count;
    }
    return (result + value.toString(radix)).slice(-digits);
  };

  intToHex = function(num, radix) {
    if (radix == null) {
      radix = 16;
    }
    if (num < 0) {
      return (num + MaxUInt32Count).toString(radix);
    } else {
      return num.toString(radix);
    }
  };

  BytewiseCodec = (function() {
    var baseWidth, buffer, bufferEncoding, decode, decodeArray, decodeBuffer, decodeDouble, decodeFunction, decodeInt32, decodeNumber, decodeObject, decodeRegExp, encode, encodeArray, encodeDouble, encodeInt32, encodeObject, getBuffer, ignoreCircular, integerBase, isCircular, sortArray, sortObject, xorBuffer;

    function BytewiseCodec() {}

    register(BytewiseCodec, Codec);

    aliases(BytewiseCodec, 'index');

    getBuffer = function(length, reuseable) {
      if (reuseable == null) {
        reuseable = true;
      }
      if (reuseable) {
        return Codec.getBuffer(length);
      } else {
        return new Buffer(length);
      }
    };

    buffer = getBuffer(8);

    decodeFunction = void 0;

    bufferEncoding = "hex";

    integerBase = 16;

    baseWidth = intToHex(MaxUInt32, integerBase).length;

    ignoreCircular = true;

    sortArray = false;

    sortObject = true;


    /*
    xorNumber = (value)->
      result = ""
      #the value must be less than the MaxIntValue, but xor ...
      x = value ^ MaxUInt32 # xor the js bitwise operation is sign 32bit integer.
      result = intToHex(x) + result
      while (value = (value / MaxUInt32Count) >> 0) > 0
        x = value ^ MaxUInt32 # xor the js bitwise operation is sign 32bit integer.
        result = intToHex(x) + result
      result
     */

    xorBuffer = function(buf, offset, end) {
      var i;
      if (offset == null) {
        offset = 0;
      }
      if (end == null) {
        end = 8;
      }
      if (end > buf.length) {
        end = buf.length;
      }
      i = offset;
      while (i < end) {
        buf[i] = buf[i] ^ 0xff;
        i++;
      }
    };

    encodeInt32 = function(value) {
      return toFixedInt(value, baseWidth, integerBase);
    };

    encodeDouble = function(value) {
      if (value >= 0) {
        buffer.writeDoubleBE(value, 0);
      } else {
        buffer.writeDoubleBE(-value, 0);
        xorBuffer(buffer);
      }
      return buffer.toString(bufferEncoding, 0, 8);
    };

    isCircular = (function() {
      var visited;
      visited = [];
      return function(obj, clearVisited) {
        var result;
        if (clearVisited === true) {
          visited = [];
          if (isObject(obj)) {
            visited.push(obj);
          }
          result = false;
        } else if (obj) {
          result = visited.indexOf(obj) >= 0;
          if (!result && isObject(obj)) {
            visited.push(obj);
          }
        }
        return result;
      };
    })();

    encodeArray = function(arr) {
      var i, item, result, _i, _len;
      result = [];
      for (i = _i = 0, _len = arr.length; _i < _len; i = ++_i) {
        item = arr[i];
        if (isCircular(item)) {
          if (ignoreCircular) {
            continue;
          } else {
            throw new InvalidFormatError("Circular element found.");
          }
        }
        result.push(encode(item));
      }
      if (sortArray) {
        result = result.sort();
      }
      return '[' + result.join(',') + ']';
    };

    encodeObject = function(obj) {
      var k, keys, result, v, _i, _len;
      result = [];
      keys = Object.keys(obj);
      if (sortObject) {
        keys = keys.sort();
      }
      for (_i = 0, _len = keys.length; _i < _len; _i++) {
        k = keys[_i];
        v = obj[k];
        if (isCircular(v)) {
          if (ignoreCircular) {
            continue;
          } else {
            throw new InvalidFormatError("Circular element found.");
          }
        }
        result.push(escapeString(k, "%:,") + ':' + encode(v));
      }
      return '{' + result.join(',') + '}';
    };

    encode = function(data) {
      var type, value;
      if (data === void 0) {
        return UNDEFINED;
      }
      if (data === null) {
        return NULL;
      }
      value = (data != null) && data.valueOf ? data.valueOf() : data;
      if (value !== value) {
        if (data instanceof Date) {
          throw new InvalidArgumentError('invalid date value');
        }
        throw new InvalidArgumentError('NaN value is not permitted');
      }
      if (value === false) {
        return FALSE;
      }
      if (value === true) {
        return TRUE;
      }
      if (isString(data)) {
        return JSON.stringify(escapeString(value, "%,:"));
      }
      if (isDate(data)) {
        if (Object.is(value, -0)) {
          value = 0;
        }
        type = value < 0 ? DATE_PRE_EPOCH : DATE_POST_EPOCH;
        return type + encodeDouble(value);
      }
      if (isNumber(data)) {
        if (value === Number.NEGATIVE_INFINITY) {
          return NUMBER_NEGATIVE_INFINITY;
        }
        if (value === Number.POSITIVE_INFINITY) {
          return NUMBER_POSITIVE_INFINITY;
        }
        if (Object.is(value, -0)) {
          value = 0;
        }
        if (value === (value | 0) && Math.abs(value) <= MaxUInt32Count) {
          type = value < 0 ? NEGATIVE_INT : POSITIVE_INT;
          return type + encodeInt32(value);
        } else {
          type = value < 0 ? NEGATIVE_FLOAT : POSITIVE_FLOAT;
          return type + encodeDouble(value);
        }
      }
      if (isRegExp(data)) {
        return REGEXP + value.toString();
      }
      if (isBuffer(data)) {
        return BUFFER + data.toString(bufferEncoding);
      }
      if (isArray(data)) {
        return encodeArray(value);
      }
      if (isFunction(data)) {
        return value.toString();
      }
      if (isObject(data) && !(data instanceof Error)) {
        return encodeObject(value);
      }
      throw new NotSupportedError('the value type is not supported:' + data);
    };

    decodeRegExp = function(str) {
      var flags, lastIndex, pattern;
      lastIndex = str.length - 1;
      if (str[lastIndex] !== '/') {
        lastIndex = str.lastIndexOf('/');
        flags = str.subString(lastIndex + 1);
      }
      pattern = str.subString(1, lastIndex);
      return new RegExp(pattern, flags);
    };

    decodeArray = function(data) {
      var item, _i, _len, _ref, _results;
      _ref = data.split(",");
      _results = [];
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        item = _ref[_i];
        _results.push(decode(item));
      }
      return _results;
    };

    decodeObject = function(data) {
      var item, key, result, value, _i, _len, _ref, _ref1;
      result = {};
      _ref = data.split(",");
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        item = _ref[_i];
        _ref1 = item.split(":"), key = _ref1[0], value = _ref1[1];
        key = unescapeString(key);
        value = decode(value);
        result[key] = value;
      }
      return result;
    };

    decodeBuffer = function(data) {
      var buf, len;
      buf = getBuffer(data.length, false);
      len = buf.write(data, 0, bufferEncoding);
      return buf.slice(0, len);
    };

    decodeNumber = function(data) {
      var c;
      c = data[0];
      data = data.slice(1);
      if (c === INTEGER) {
        return decodeInt32(data);
      } else if (c === FLOAT) {
        return decodeDouble(data);
      } else {
        throw new InvalidFormatError("unknown number type flag:", c);
      }
    };

    decodeInt32 = function(data) {
      var c, result;
      c = data[0];
      data = data.slice(1);
      result = parseInt(data, integerBase);
      if (c === NEGATIVE) {
        return -((result ^ MaxUInt32) + 1);
      } else if (c === POSITIVE) {
        return result;
      } else {
        throw new InvalidFormatError("unknown sign flag:", c);
      }
    };

    decodeDouble = function(data) {
      var c;
      c = data[0];
      data = data.slice(1);
      buffer.write(data, 0, 8, bufferEncoding);
      if (c === NEGATIVE) {
        xorBuffer(buffer, 0, 8);
        return -buffer.readDoubleBE(0);
      } else if (c === POSITIVE) {
        return buffer.readDoubleBE(0);
      } else {
        throw new InvalidFormatError("unknown sign flag:", c);
      }
    };

    decode = function(data) {
      if (isString(data) && data.length > 0) {
        if (data === NUMBER_NEGATIVE_INFINITY) {
          return Number.NEGATIVE_INFINITY;
        }
        if (data === NUMBER_POSITIVE_INFINITY) {
          return Number.POSITIVE_INFINITY;
        }
        switch (data[0]) {
          case STRING:
            return JSON.parse(unescapeString(data));
          case TRUE:
            return true;
          case FALSE:
            return false;
          case NULL:
            return null;
          case UNDEFINED:
            return void 0;
          case NUMBER:
            return decodeNumber(data.slice(1));
          case DATE:
            return new Date(decodeDouble(data.slice(1)));
          case ARRAY:
            return decodeArray(data.slice(1, data.length - 1));
          case OBJECT:
            return decodeObject(data.slice(1, data.length - 1));
          case BUFFER:
            return decodeBuffer(data.slice(1));
          case REGEXP:
            return decodeRegExp(data.slice(1));
          case FUNCTION:
            if (decodeFunction) {
              return decodeFunction(data);
            } else {
              return data;
            }
          default:
            return data;
        }
      } else {
        return data;
      }
    };

    BytewiseCodec.prototype._encodeString = function(data) {
      var i;
      isCircular(null, true);
      i = this.options.sortObject;
      sortObject = i != null ? i : true;
      i = this.options.sortArray;
      sortArray = i != null ? i : false;
      return encode(data);
    };

    BytewiseCodec.prototype._decodeString = decode;

    BytewiseCodec.prototype.config = function(conf) {
      if (conf) {
        decodeFunction = conf.decodeFunction;
        if (conf.bufferEncoding) {
          bufferEncoding = conf.bufferEncoding;
        }
        ignoreCircular = conf.ignoreCircular;
        if (isFunction(conf.getBuffer)) {
          getBuffer = conf.getBuffer;
          buffer = getBuffer(8);
        }
        if (conf.integerBase >= 2 && conf.integerBase <= 36) {
          integerBase = conf.integerBase;
          baseWidth = intToHex(MaxUInt32, integerBase).length;
        }
        return this;
      } else {
        return {
          decodeFunction: decodeFunction,
          bufferEncoding: bufferEncoding,
          integerBase: integerBase
        };
      }
    };

    return BytewiseCodec;

  })();

}).call(this);

//# sourceMappingURL=bytewise-codec.js.map
